package de.zebee.coding;

import java.lang.ref.SoftReference;
import java.util.HashMap;


/**
 * cyclic redundancy code
 * @author Sebastian Gesemann
 */
public class CRC {


	public static class CRCSpec {
		public final int bits;
		public final int poly;
		public CRCSpec(int bits, int poly) {
			this.bits = bits;
			this.poly = poly;
		}
	}


	/*
	Standard Polynomials
	--------------------
	CRC-12:		X^12+X^11+X^3+X^2+X+1 
	CRC-16:		X^16+X^15+X^2+1 
	CRC-CCITT:	X^16+X^12+X^5+1 
	CRC-32:		X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X+1
	*/


	public static final int MIN_BITS = 8;
	public static final int MAX_BITS = 32;

	public static final CRCSpec CRC_12 = new CRCSpec(12,0x80F);
	public static final CRCSpec CRC_16 = new CRCSpec(16,0x8005);
	public static final CRCSpec CRC_CCITT16 = new CRCSpec(16,0x1021);
	public static final CRCSpec CRC_32 = new CRCSpec(32,0x04C11DB7);


	// look-up-table cache
	private static final HashMap luts[] = new HashMap[MAX_BITS-MIN_BITS+1];


	int bits;		// size in bits of the register
	int upperdown;	// = bits - 8
	int mask;		// = (1 << bits) - 1
	int poly;		// binary representation of the generator polynomial
	int[] lut;		// look-up-table

	int register = 0;


	/**
	 * Contructs a new CRC object with a register and generator
	 * polynomial specified by the CRCSpec object.
	 * @param crcSpec
	 */
	public CRC(CRCSpec crcSpec) {
		this(crcSpec.bits, crcSpec.poly);
	}

	/**
	 * Contructs a new CRC object with a register size of
	 * 'bits' bits and a generator polynomial 'poly'. The
	 * register will be initialized with zero.
	 * @param bits
	 * @param poly
	 */
	public CRC(int bits, int poly) {
		this.bits = bits;
		this.upperdown = bits-8;
		this.mask = ((1 << upperdown) << 8) - 1;
		this.poly = poly & mask;
		getLUT();
	}

	/**
	 * returns the size of the register in bits
	 * @return int
	 */
	public int getRegisterSize() {
		return bits;
	}

	/**
	 * returns the generator polynomial (without X^'bits')
	 * @return int
	 */
	public int getGeneratorPolynomial() {
		return poly;
	}

	/**
	 * sets the register's state
	 * @param r
	 */
	public void setRegisterState(int r) {
		register = r;
	}

	/**
	 * processes the bits of a byte (from MSB to LSB)
	 * @param bite
	 */
	public void process(int bite) {
		register = (register << 8) ^ lut[((register >> upperdown) ^ bite) & 0xFF];
	}

	/**
	 * processes 'len' bytes starting at 'offset'
	 * @param data
	 * @param offset
	 * @param len
	 */
	public void process(byte[] data, int offset, int len) {
		while (len-->0) {
			register = (register << 8) ^ lut[((register >> upperdown) ^ data[offset++]) & 0xFF];
		}
	}

	/**
	 * processes all bytes of the byte array
	 * @param data
	 */
	public void process(byte[] data) {
		process(data,0,data.length);
	}

	/**
	 * returns the register's state;
	 * @return int
	 */
	public int getRegisterState() {
		return register & mask;
	}

	/**
	 * get the correct LookUpTable for the current CRC-Code
	 * by searching it in the LUT cache. It will be created by
	 * invoking createNewLUT() and added to the cache, if it
	 * hasn't been created yet.
	 */
	private void getLUT() {
		int tableIdx = bits-MIN_BITS;
		if ((tableIdx<0) || (bits>MAX_BITS)) throw new RuntimeException("bit count parameter out of range");
		HashMap hm = luts[tableIdx];
		if (hm==null) {
			hm = luts[tableIdx] = new HashMap();
		}
		Integer key = new Integer(poly);
		SoftReference lutRef;
		if (hm.containsKey(key)) {
			lutRef = (SoftReference)hm.get(key);
			lut = (int[])lutRef.get();
			if (lut!=null) return;
			hm.remove(key); // remove null-pointer reference
		}
		createNewLUT();
		lutRef = new SoftReference(lut);
		hm.put(key,lutRef);
	}

	/**
	 * Actually compute the LUT.
	 */
	private void createNewLUT() {
		lut = new int[0x100];
		for (int z=0; z<0x100; z++) {
			int tmp = 0;
			for (int s=0; s<8; s++) {
				int x = (tmp >> upperdown) ^ (z << s);
				tmp <<= 1;
				if ((x & 0x80) != 0) tmp ^= poly;
			}
			lut[z] = tmp & mask;
		}
	}

}
